"""
Main window widgets for running applications
"""

from imgui_bundle import imgui, implot, implot3d, imgui_md, hello_imgui
from ymery.frontend.widget import Widget
from ymery.frontend.composite import Composite
from ymery.decorators import widget
from ymery.result import Result, Ok


@widget
class HelloImguiMainWindow(Composite):
    """
    Main window using hello_imgui.run() with docking support

    Parameters:
        window-title: Window title (default: "Ymery App")
        window-size: Window size as [width, height] (default: [1200, 800])
        fps-idle: FPS when idle (default: 0)

    Body can contain:
        - DockingSplit widgets (define dock space layout)
        - DockableWindow widgets (dockable windows)
        - Other widgets (rendered in main dock space if no docking)
    """

    def init(self) -> Result[None]:
        """Initialize the main window"""
        res = super().init()
        if not res:
            return Result.error("HelloImguiMainWindow: failed to initialize Widget", res)

        # Initialize ImPlot context
        if not implot.get_current_context():
            implot.create_context()

        # Initialize ImPlot3D context
        if not implot3d.get_current_context():
            implot3d.create_context()

        # Initialize Markdown
        imgui_md.initialize_markdown()

        return Ok(None)


    def _render_menu_bar(self):
        """Render the menu bar"""
        if self._menu_bar_widget:
            self._handle_error(self._menu_bar_widget.render())

    def _main_loop(self):
        """Main rendering loop called by hello_imgui"""
        imgui.text("Main loop is running!")
        # Render non-docking widgets (regular widgets in main dock space)
        for widget in self._non_docking_widgets:
            self._handle_error(widget.render())

        self._render_errors()

    def _resolve_constant(self, field_name: str, value: str):
        """Convert YAML constant to Python constant

        Examples:
        - field: default-imgui-window-type, value: provide-full-screen-dock-space
          → hello_imgui.DefaultImGuiWindowType.provide_full_screen_dock_space
        - field: direction, value: down
          → imgui.Dir.down
        """
        # Replace hyphens with underscores
        py_value = value.replace('-', '_')

        # Map field names to namespace
        constant_map = {
            'default-imgui-window-type': 'hello_imgui.DefaultImGuiWindowType',
            'direction': 'imgui.Dir',
        }

        if field_name in constant_map:
            namespace = constant_map[field_name]
            return eval(f"{namespace}.{py_value}")

        return value

    def _process_params_dict(self, d: dict, parent_key: str = "") -> dict:
        """Recursively process dict, resolving constants"""
        result = {}
        for key, value in d.items():
            full_key = f"{parent_key}.{key}" if parent_key else key

            if isinstance(value, dict):
                result[key.replace('-', '_')] = self._process_params_dict(value, full_key)
            elif isinstance(value, str):
                # Try to resolve as constant
                result[key.replace('-', '_')] = self._resolve_constant(key, value)
            elif isinstance(value, list):
                result[key.replace('-', '_')] = value
            else:
                result[key.replace('-', '_')] = value

        return result

    def _read_runner_params_metadata(self) -> dict:
        """Read runner-params metadata and convert to Python dict"""
        res = self._data_bag.get("runner-params", {})
        if not res:
            return {}

        params_dict = res.unwrapped
        # Recursively process dict to resolve constants
        return self._process_params_dict(params_dict)

    def run(self) -> Result[None]:
        """Run the application - this is called by app.py"""
        # Initialize children
        res = self.children
        if not res:
            return 1
        children = res.unwrapped

        self._non_docking_widgets = []
        self._menu_bar_widget = None
        # Extract children by type
        docking_splits = []
        dockable_windows = []
        menu_widget = None
        app_menu_items_widget = None

        for child in children:
            if isinstance(child, DockingSplit):
                docking_splits.append(child)
            elif isinstance(child, DockableWindow):
                dockable_windows.append(child)
            elif child.__class__.__name__ ==  "MainMenuBar":
                if self._menu_bar_widget is not None:
                    self._handle_error(Result.error("HelloImguiMainWindow: multiple MainMenuBar widgets found, only one allowed"))
                else:
                    self._menu_bar_widget = child
            elif child.__class__.__name__ == "HelloImguiMenu":
                if menu_widget is not None:
                    self._handle_error(Result.error("HelloImguiMainWindow: multiple HelloImguiMenu widgets found, only one allowed"))
                else:
                    menu_widget = child
            elif child.__class__.__name__ == "HelloImguiAppMenuItems":
                if app_menu_items_widget is not None:
                    self._handle_error(Result.error("HelloImguiMainWindow: multiple HelloImguiAppMenuItems widgets found, only one allowed"))
                else:
                    app_menu_items_widget = child
            else:
                # Other widgets render in main loop
                self._non_docking_widgets.append(child)

        # Setup hello_imgui RunnerParams
        runner_params = hello_imgui.RunnerParams()

        # Read and apply runner-params metadata
        params_dict = self._read_runner_params_metadata()

        # Apply app_window_params
        if 'app_window_params' in params_dict:
            for key, val in params_dict['app_window_params'].items():
                if key == 'window_geometry' and isinstance(val, dict):
                    for gkey, gval in val.items():
                        setattr(runner_params.app_window_params.window_geometry, gkey, gval)
                else:
                    setattr(runner_params.app_window_params, key, val)

        # Apply imgui_window_params
        if 'imgui_window_params' in params_dict:
            for key, val in params_dict['imgui_window_params'].items():
                setattr(runner_params.imgui_window_params, key, val)

        # Fallback to old fields if not in runner-params
        res = self._data_bag.get("label", self.uid)
        if not res:
            self._handle_error(Result.error("HelloImguiMainWindow: failed to get label", res))
            label = "error: LABEL NOT AVAILABLE"
        else:
            label = res.unwrapped
        if 'app_window_params' not in params_dict:
            runner_params.app_window_params.window_title = label

        res = self._data_bag.get("window-size", [1200, 800])
        if not res:
            self._handle_error(Result.error("HelloImguiMainWindow: failed to get window-size", res))
            size_list = [1200, 800]
        else:
            size_list = res.unwrapped

        if 'app_window_params' not in params_dict:
            runner_params.app_window_params.window_geometry.size = (size_list[0], size_list[1])

        # If we have dockable windows or docking splits, enable docking
        if dockable_windows or docking_splits:
            runner_params.imgui_window_params.default_imgui_window_type = (
                hello_imgui.DefaultImGuiWindowType.provide_full_screen_dock_space
            )

            # Setup docking params - create complete DockingParams object
            docking_params = hello_imgui.DockingParams()

            # Set layout condition to always apply the layout
            docking_params.layout_condition = hello_imgui.DockingLayoutCondition.application_start

            # Add docking splits
            docking_params.docking_splits = [split_widget.docking_split for split_widget in docking_splits]

            # Add dockable windows
            docking_params.dockable_windows = [window_widget.dockable_window for window_widget in dockable_windows]

            # Assign the complete DockingParams to runner_params
            runner_params.docking_params = docking_params

        # Set main GUI function
        runner_params.callbacks.show_gui = self._main_loop

        # Load markdown fonts
        runner_params.callbacks.load_additional_fonts = imgui_md.get_font_loader_function()

        # Set menu bar callback if menu bar widget exists
        if self._menu_bar_widget:
            runner_params.imgui_window_params.show_menu_bar = True
            runner_params.callbacks.show_menus = self._render_menu_bar

        # Set HelloImGui menu callbacks
        if menu_widget:
            runner_params.imgui_window_params.show_menu_bar = True
            runner_params.callbacks.show_menus = lambda: menu_widget.render()

        if app_menu_items_widget:
            runner_params.callbacks.show_app_menu_items = lambda: app_menu_items_widget.render()

        res = self._data_bag.get("fps-idle", 0)
        if not res:
            self._handle_error(Result.error("HelloImguiMainWindow: failed to get 'fps-idle'"))
            idle = 0
        else:
            idle = res.unwrapped
        runner_params.fps_idling.fps_idle = idle

        # Run application
        hello_imgui.run(runner_params)

        return Ok[None]

    def _pre_render_head(self) -> Result[None]:
        """HelloImguiMainWindow doesn't render in normal flow"""
        return Ok(None)

    def _post_render_head(self) -> Result[None]:
        """No cleanup needed"""
        return Ok(None)

"""
HelloImGui-specific widgets for menu bar integration
"""


@widget
class HelloImguiMenu(Composite):
    """HelloImGui menu bar contents widget

    Renders children inside HelloImGui's menu bar context.
    HelloImGui handles begin_main_menu_bar() / end_main_menu_bar() automatically.
    This widget just renders its Menu children.
    """
    pass  # No overrides needed - Composite handles rendering children


@widget
class HelloImguiAppMenuItems(Composite):
    """HelloImGui app menu items widget

    Adds custom items to the HelloImGui App menu.
    Children should be menu-item widgets.
    """
    pass  # No overrides needed - Composite handles rendering children



"""
Docking widgets for hello_imgui docking system
"""



@widget
class DockableWindow(Widget):
    """
    Dockable window widget - creates a window that can be docked in the hello_imgui docking system

    Parameters:
        label: Window title
        dock-space-name: Name of the dock space to dock to
        include-in-view-menu: Whether to include in View menu (default: True)
        remember-is-visible: Whether to remember visibility state (default: True)
        is-visible: Initial visibility state (default: True)
        can-be-closed: Whether window can be closed (default: True)
    """

    def init(self) -> Result[None]:
        """Initialize the dockable window"""
        self._dockable_window = None
        res = super().init()
        if not res:
            return Result.error("DockableWindow: failed to initialize Widget", res)

        # Create hello_imgui.DockableWindow instance
        self._dockable_window = hello_imgui.DockableWindow()

        # Get label
        res = self._data_bag.get("label")
        if not res:
            return Result.error("DockableWindow: failed to get label", res)
        self._dockable_window.label = res.unwrapped

        # Get dock_space_name (optional)
        res = self._data_bag.get("dock-space-name", "MainDockSpace")
        if res:
            self._dockable_window.dock_space_name = res.unwrapped

        # Get optional parameters
        res = self._data_bag.get("include-in-view-menu", True)
        if res:
            self._dockable_window.include_in_view_menu = res.unwrapped

        res = self._data_bag.get("remember-is-visible", True)
        if res:
            self._dockable_window.remember_is_visible = res.unwrapped

        res = self._data_bag.get("is-visible", True)
        if res:
            self._dockable_window.is_visible = res.unwrapped

        res = self._data_bag.get("can-be-closed", True)
        if res:
            self._dockable_window.can_be_closed = res.unwrapped

        # Set gui_function to render the body
        self._dockable_window.gui_function = self._render_dockable

        return Ok(None)

    def _render_dockable(self):
        res = self.render()
        if not res:
            # TODO use the render_error from widget.py to display the error!
            print("Failed to render dockable window")
            import yaml
            print(yaml.dump(res.as_tree))
            import sys
            sys.exit(-1)

    def _pre_render_head(self) -> Result[None]:
        """DockableWindow doesn't render in the normal flow - it's handled by hello_imgui"""
        # Mark as activated so body gets created and rendered
        self._is_body_activated = True
        return Ok(None)

    def _post_render_head(self) -> Result[None]:
        """No cleanup needed"""
        return Ok(None)


    @property
    def dockable_window(self):
        """Get the hello_imgui.DockableWindow instance"""
        return self._dockable_window

    def dispose(self) -> Result[None]:
        """Cleanup"""
        return Ok(None)


@widget
class DockingSplit(Widget):
    """
    Docking split widget - defines how to split a dock space into multiple regions

    Parameters:
        initial-dock: Name of the initial dock space to split
        new-dock: Name of the new dock space to create
        direction: Direction to split (up, down, left, right)
        ratio: Ratio of the split (0.0 to 1.0)
        body: Optional list of DockableWindow widgets that will automatically use new-dock as their dock-space-name
    """

    def init(self) -> Result[None]:
        """Initialize the docking split"""
        res = super().init()
        if not res:
            return Result.error("DockingSplit: failed to initialize Widget", res)

        # Create hello_imgui.DockingSplit instance
        self._docking_split = hello_imgui.DockingSplit()

        # Get initial-dock
        res = self._data_bag.get("initial-dock")
        if not res:
            return Result.error("DockingSplit: failed to get initial-dock", res)
        self._docking_split.initial_dock = res.unwrapped

        # Get new-dock
        res = self._data_bag.get("new-dock")
        if not res:
            return Result.error("DockingSplit: failed to get new-dock", res)
        self._docking_split.new_dock = res.unwrapped

        # Get direction
        res = self._data_bag.get("direction", "down")
        if not res:
            return Result.error("DockingSplit: failed to get direction", res)

        direction_str = res.unwrapped
        direction_map = {
            "up": imgui.Dir_.up,
            "down": imgui.Dir_.down,
            "left": imgui.Dir_.left,
            "right": imgui.Dir_.right
        }

        if direction_str not in direction_map:
            return Result.error(f"DockingSplit: invalid direction '{direction_str}', must be one of: up, down, left, right")

        self._docking_split.direction = direction_map[direction_str]

        # Get ratio
        res = self._data_bag.get("ratio", 0.5)
        if not res:
            return Result.error("DockingSplit: failed to get ratio", res)
        self._docking_split.ratio = float(res.unwrapped)

        return Ok(None)


    def render(self) -> Result[None]:
        """DockingSplit doesn't render - it's configuration only"""
        return Ok(None)

    @property
    def docking_split(self):
        """Get the hello_imgui.DockingSplit instance"""
        return self._docking_split

    def dispose(self) -> Result[None]:
        """Cleanup"""
        return Ok(None)



from ymery.stringcase import spinalcase

_pending_widgets = {}
def widget(name_or_cls=None):
    def decorator(cls):
        key = name_or_cls if isinstance(name_or_cls, str) else spinalcase(cls.__name__)
        _pending_widgets[key] = cls
        return cls
    if name_or_cls is None or isinstance(name_or_cls, str):
        return decorator
    _pending_widgets[spinalcase(name_or_cls.__name__)] = name_or_cls
    return name_or_cls


_pending_device_managers = {}
def device_manager(name_or_cls=None):
    def decorator(cls):
        key = name_or_cls if isinstance(name_or_cls, str) else spinalcase(cls.__name__)
        _pending_device_managers[key] = cls
        return cls
    if name_or_cls is None or isinstance(name_or_cls, str):
        return decorator
    _pending_device_managers[spinalcase(name_or_cls.__name__)] = name_or_cls
    return name_or_cls

_pending_devices = {}
def device(name_or_cls=None):
    def decorator(cls):
        key = name_or_cls if isinstance(name_or_cls, str) else spinalcase(cls.__name__)
        _pending_devices[key] = cls
        return cls
    if name_or_cls is None or isinstance(name_or_cls, str):
        return decorator
    _pending_devices[spinalcase(name_or_cls.__name__)] = name_or_cls
    return name_or_cls

_pending_tree_likes = {}
def tree_like(name_or_cls=None):
    def decorator(cls):
        key = name_or_cls if isinstance(name_or_cls, str) else spinalcase(cls.__name__)
        _pending_tree_likes[key] = cls
        return cls
    if name_or_cls is None or isinstance(name_or_cls, str):
        return decorator
    _pending_tree_likes[spinalcase(name_or_cls.__name__)] = name_or_cls
    return name_or_cls



"""
Generic application runner for ymery
"""

import sys
import logging
import click
from pathlib import Path
from imgui_bundle import imgui, immapp
from ymery.lang import Lang
from ymery.frontend.widget_factory import WidgetFactory
from ymery.plugin_manager import PluginManager
from ymery.dispatcher import Dispatcher
from ymery.types import DataPath
from ymery.result import Result, Ok
from ymery.frontend.widget import render_error
from ymery.logging import setup_logging, get_ring_buffer

import time

init_time = time.time()


def show_if_error(err) -> Result[None]:
    """Run a simple imgui loop to display an error"""
    if err:
        return err
    error_tree = err.as_tree if hasattr(err, 'as_tree') else {"error": str(err)}
    logging.error(error_tree)

    def gui_loop():
        imgui.text("The app could not be initialized due to the errors below. Click 'exit' to terminate the app")
        if imgui.button("exit"):
            sys.exit(-1)
        render_error(error_tree)

    immapp.run(
        gui_function=gui_loop,
        window_title="Ymery - Error",
        window_size=(800, 600)
    )
    return Ok(None)


@click.command()
@click.option('--layouts-path', '-p',
              envvar='YMERY_LAYOUTS_PATH',
              type=str,
              multiple=True,
              help='Colon-separated list of directories to search for layout modules (can be specified multiple times)')
@click.option('--layouts-url', '-u',
              envvar='YMERY_LAYOUTS_URL',
              type=str,
              multiple=True,
              help='URL to download layouts from (can be specified multiple times, e.g., GitHub raw URLs)')
@click.option('--plugins-path',
              envvar='YMERY_PLUGINS_PATH',
              type=str,
              help='Colon-separated list of directories to search for provider modules')
@click.option('--widgets-path',
              envvar='YMERY_WIDGETS_PATH',
              type=str,
              help='Colon-separated list of directories to search for widget modules')
@click.option('--main', '-m',
              required=True,
              type=str,
              help='Name of the main module to load')
@click.option('--log-level', '-l',
              envvar='YMERY_LOG_LEVEL',
              type=click.Choice(['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', '10', '20', '30', '40', '50']),
              default='INFO',
              help='Log level (DEBUG=10, INFO=20, WARNING=30, ERROR=40, CRITICAL=50)')
@click.option('--log-file', '-f',
              envvar='YMERY_LOG_FILE',
              type=str,
              default=None,
              help='Log file path (default: stdout)')

def main(layouts_path, layouts_url, plugins_path, widgets_path, main, log_level, log_file):
    """Ymery application runner"""
    global dispatcher, kernel, data_tree

    # Setup logging first
    level = getattr(logging, log_level) if log_level.isalpha() else int(log_level)
    setup_logging(level=level, log_file=log_file)

    # Combine layouts_path and layouts_url in order of specification
    # Both can be specified multiple times, precedence is order of appearance
    layouts_paths = []

    # Process --layouts-path (can be multiple, each colon-separated)
    if layouts_path:
        for path_spec in layouts_path:
            layouts_paths.extend(path_spec.split(':'))

    # Process --layouts-url (can be multiple)
    if layouts_url:
        layouts_paths.extend(layouts_url)

    # Default to current directory if nothing specified
    if not layouts_paths:
        layouts_paths = ['.']

    # Create Lang and load modules
    lang = show_if_error(Lang.create(layouts_paths=layouts_paths, main=main)).unwrapped

    # Create Dispatcher
    dispatcher = show_if_error(Dispatcher.create()).unwrapped

    # Build default plugin paths
    ymery_dir = Path(__file__).parent
    default_plugins_path = str(ymery_dir / "plugins" / "frontend") + ":" + str(ymery_dir / "plugins" / "backend")
    if plugins_path:
        plugins_path = default_plugins_path + ":" + plugins_path
    else:
        plugins_path = default_plugins_path

    # Create PluginManager
    plugin_manager = show_if_error(PluginManager.create(plugins_path=plugins_path)).unwrapped

    # Instantiate all data trees from lang.data_definitions
    data_trees = {}
    data_definitions = lang.data_definitions

    for data_name, data_def in data_definitions.items():
        # Get the type (tree-like class name)
        type_name = data_def.get('type')
        if not type_name:
            show_if_error(Result.error(f"Data '{data_name}' missing 'type' field"))
            return 1

        # Get the class from plugin_manager
        res = plugin_manager.get_metadata(DataPath(f"/tree-like/{type_name}"))
        if not res:
            show_if_error(Result.error(f"TreeLike type '{type_name}' not found for data '{data_name}'", res))
            return 1

        tree_like_class = res.unwrapped.get("class")
        if not tree_like_class:
            show_if_error(Result.error(f"No class found for TreeLike type '{type_name}'"))
            return 1

        # Get raw_arg if present
        raw_arg = data_def.get('arg')

        # Instantiate with standard args
        try:
            instance = tree_like_class(dispatcher, plugin_manager, raw_arg)
        except Exception as e:
            show_if_error(Result.error(f"Failed to instantiate '{type_name}' for data '{data_name}'", e))
            return 1

        # Initialize the instance
        res = instance.init()
        if not res:
            show_if_error(Result.error(f"Failed to init '{type_name}' for data '{data_name}'", res))
            return 1

        data_trees[data_name] = instance

    # Create WidgetFactory with all data trees
    widget_factory = show_if_error(WidgetFactory.create(dispatcher, plugin_manager, widget_definitions=lang.widget_definitions, data_trees=data_trees)).unwrapped

    # Get app config
    app_config = lang.app_config
    if not app_config:
        show_if_error(Result.error("No app configuration found"))
        return 1

    widget_name = app_config.get('widget')
    data_name = app_config.get('data')

    if not widget_name:
        show_if_error(Result.error("app.widget not specified"))
        return 1

    if not data_name:
        show_if_error(Result.error("app.data not specified"))
        return 1

    # Prepend main module namespace if widget_name doesn't have one
    if '.' not in widget_name:
        widget_name = f"{main}.{widget_name}"

    # Get the main data tree
    if data_name not in data_trees:
        show_if_error(Result.error(f"Data '{data_name}' not found in data definitions"))
        return 1

    data_tree = data_trees[data_name]

    # Create main widget
    main_widget = show_if_error(widget_factory.create_widget(widget_name, data_tree, DataPath("/"))).unwrapped

    # Run application - the main widget handles everything
    return main_widget.run()


if __name__ == '__main__':
    if sys.platform == 'emscripten':
        # In Pyodide/Emscripten, don't let Click call sys.exit()
        main(standalone_mode=False)
    else:
        main()



import logging
import time
from collections import deque
from dataclasses import dataclass, field
from typing import Optional, List, Any, Dict

# Characters for UID generation (letters + digits)
_UID_CHARS = "0123456789abcdefghijklmnopqrstuvwxyz"


def _hash_to_uid(h: int, length: int = 10) -> str:
    """Convert a hash to a short UID string (letters + digits)."""
    h = abs(h)
    base = len(_UID_CHARS)
    result = []
    for _ in range(length):
        result.append(_UID_CHARS[h % base])
        h //= base
    return ''.join(result)


def _compute_msg_hash(record: logging.LogRecord) -> int:
    """Compute hash for deduplication based on message content."""
    msg = record.msg
    if hasattr(msg, '__hash__') and callable(msg.__hash__):
        try:
            msg_hash = hash(msg)
        except TypeError:
            msg_hash = hash(str(msg))
    else:
        msg_hash = hash(str(msg))
    return hash((record.levelno, record.name, msg_hash))


@dataclass
class _DedupEntry:
    """Tracks a deduplicated message."""
    uid: str
    count: int = 1
    first_time: float = field(default_factory=time.time)
    last_summary_time: float = field(default_factory=time.time)
    level: int = logging.ERROR
    name: str = "root"


class DedupFilter(logging.Filter):
    """
    Filter that deduplicates log messages.

    - First occurrence: adds UID, allows through
    - Repeated: suppresses, but emits summary every dedup_window seconds
    """

    def __init__(self, dedup_window: float = 1.0):
        super().__init__()
        self.dedup_window = dedup_window
        self._cache: Dict[int, _DedupEntry] = {}

    def filter(self, record: logging.LogRecord) -> bool:
        msg_hash = _compute_msg_hash(record)
        current_time = time.time()

        if msg_hash in self._cache:
            entry = self._cache[msg_hash]
            entry.count += 1

            # Check if it's time to emit summary
            if current_time - entry.last_summary_time >= self.dedup_window:
                elapsed = int(current_time - entry.first_time)
                # Modify record to be a summary
                record.msg = f"error '{entry.uid}' repeated {entry.count} times the last {elapsed} seconds"
                record.args = ()
                record.uid = entry.uid
                entry.last_summary_time = current_time
                return True

            # Suppress duplicate
            return False
        else:
            # New unique message
            uid = _hash_to_uid(msg_hash)
            self._cache[msg_hash] = _DedupEntry(
                uid=uid,
                count=1,
                first_time=current_time,
                last_summary_time=current_time,
                level=record.levelno,
                name=record.name
            )
            record.uid = uid
            return True


@dataclass
class LogEntry:
    """A log entry stored in the ring buffer."""
    record: logging.LogRecord
    timestamp: float = field(default_factory=time.time)

    @property
    def uid(self) -> str:
        return getattr(self.record, 'uid', '')

    @property
    def msg(self) -> Any:
        return self.record.msg

    @property
    def level(self) -> int:
        return self.record.levelno

    @property
    def level_name(self) -> str:
        return self.record.levelname

    @property
    def name(self) -> str:
        return self.record.name


class RingBufferHandler(logging.Handler):
    """Simple ring buffer handler - just stores records."""

    def __init__(self, capacity: int = 100):
        super().__init__()
        self.buffer: deque[LogEntry] = deque(maxlen=capacity)

    def emit(self, record: logging.LogRecord) -> None:
        try:
            self.buffer.append(LogEntry(record=record, timestamp=time.time()))
        except Exception:
            self.handleError(record)

    def get_entries(self) -> List[LogEntry]:
        return list(self.buffer)

    def clear(self) -> None:
        self.buffer.clear()


# Global instances
_ring_buffer_handler: Optional[RingBufferHandler] = None
_dedup_filter: Optional[DedupFilter] = None


def get_ring_buffer() -> Optional[RingBufferHandler]:
    return _ring_buffer_handler


def setup_logging(
    level: int = logging.INFO,
    log_file: Optional[str] = None,
    ring_buffer_capacity: int = 100,
    dedup_window: float = 1.0
) -> RingBufferHandler:
    """
    Configure logging with deduplication, optional file output, and ring buffer.
    """
    global _ring_buffer_handler, _dedup_filter

    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    root_logger.handlers.clear()
    root_logger.filters.clear()

    # Single dedup filter for all handlers
    _dedup_filter = DedupFilter(dedup_window=dedup_window)
    root_logger.addFilter(_dedup_filter)

    # Formatter with UID
    formatter = logging.Formatter(
        fmt='%(asctime)s [%(levelname)s] %(uid)s: %(message)s',
        datefmt='%H:%M:%S'
    )

    # Stream or file handler
    if log_file is None:
        handler = logging.StreamHandler()
    else:
        handler = logging.FileHandler(log_file)
    handler.setFormatter(formatter)
    handler.setLevel(level)
    root_logger.addHandler(handler)

    # Ring buffer handler
    _ring_buffer_handler = RingBufferHandler(capacity=ring_buffer_capacity)
    _ring_buffer_handler.setLevel(level)
    root_logger.addHandler(_ring_buffer_handler)

    return _ring_buffer_handler


def log(*args):
    """Simple log function for backward compatibility."""
    logging.info(' '.join(str(arg) for arg in args))
